%{
#include "hash.h"
#include "tunes.h"

#define YYERROR_VERBOSE
%}

%union {
  hashv_t ident;
  objptr_t val;
  char bool_v;
  char char_v;
  int int_v;
}

%token <bool_v> TOK_BOOLVAL
%token <char_v> TOK_CHARVAL
%token <int_v> TOK_INTVAL
%token <ident> TOK_IDENT
%token TOK_LPAREN
%token TOK_RPAREN
%token TOK_SEMICOLON
%token TOK_ARROW
%token TOK_EOF

%type <val> expr
%type <val> simple_expr

%start main

%%

main:
	expr				{ mainprog = $1; }
;

expr:
	simple_expr			{ $$ = $1; }
      | expr simple_expr		{ $$ = build_apply ($1, $2); }
      | expr TOK_SEMICOLON expr		{ $$ = build_progn ($1, $3); }
;

simple_expr:
	TOK_LPAREN TOK_RPAREN		{ $$ = build_unit (); }
      | TOK_BOOLVAL			{ $$ = build_bool ($1); }
      | TOK_CHARVAL			{ $$ = build_char ($1); }
      | TOK_INTVAL			{ $$ = build_int ($1); }
      | TOK_IDENT			{ $$ = build_var ($1); }
      | TOK_IDENT TOK_ARROW expr	{ $$ = build_lambda ($1, $3); }
      | TOK_LPAREN expr TOK_RPAREN	{ $$ = $2; }
;

%%

objptr_t
build_apply (objptr_t rator, objptr_t rand)
{
  objptr_t obj;

  obj = allocate ();
  core[obj].type = O_APPLY;
  core[obj].d.apply_v.rator = rator;
  core[obj].d.apply_v.rand = rand;
  return obj;
}

objptr_t
build_progn (objptr_t primus, objptr_t pares)
{
  objptr_t obj;

  obj = allocate ();
  core[obj].type = O_PROGN;
  core[obj].d.progn_v.primus = primus;
  core[obj].d.progn_v.pares = pares;
  return obj;
}

objptr_t
build_unit (void)
{
  return a_unit ();
}

objptr_t
build_bool (char bool_v)
{
  objptr_t obj;

  obj = allocate ();
  core[obj].type = O_BOOL;
  core[obj].d.bool_v = bool_v;
  return obj;
}

objptr_t
build_char (char char_v)
{
  objptr_t obj;

  obj = allocate ();
  core[obj].type = O_CHAR;
  core[obj].d.char_v = char_v;
  return obj;
}

objptr_t
build_int (int int_v)
{
  objptr_t obj;

  obj = allocate ();
  core[obj].type = O_INT;
  core[obj].d.int_v = int_v;
  return obj;
}

objptr_t
build_var (hashv_t name)
{
  objptr_t obj;

  obj = allocate ();
  core[obj].type = O_VAR;
  core[obj].d.var_v = name;
  return obj;
}

objptr_t
build_lambda (hashv_t param, objptr_t body)
{
  objptr_t obj;

  obj = allocate ();
  core[obj].type = O_LAMBDA;
  core[obj].d.lambda_v.param = param;
  core[obj].d.lambda_v.body = body;
  return obj;
}

void
yyerror (char *msg)
{
  error (msg);
}
