#ifndef _TUNES_H
#define _TUNES_H

#include "hash.h"

typedef long int objptr_t;

typedef enum {
  PF_I, PF_K, PF_S,
  PF1_K, PF1_S, PF2_S,
  PF_CALL_CC,
  PF_PRINT
} primfunc_t;

struct object_s {
  enum {
    /* Data types */
    O_UNIT,
    O_BOOL,
    O_CHAR,
    O_INT,
    O_NULL,
    O_PAIR,
    O_REF,
    /* Internal use (but reified) */
    O_PRIMFUNC,
    O_PRIMFUNC1,
    O_PRIMFUNC2,
    O_DEF,
    O_ENVNULL,
    O_ENV,
    O_USERFUNC,
    /* Continuations */
    O_CONT_FINISH,
    O_CONT_EVAL,
    O_CONT_APPLY,
    O_CONT_APPLY_1,
    /* Tasks */
    O_TASK_EVAL,
    O_TASK_APPLY,
    O_TASK_APPLY_0,
    O_TASK_APPLY_1,
    O_TASK_PROGN_0,
    /* Program types */
    O_VAR,
    O_APPLY,
    O_PROGN,
    O_QUOTE,
    O_LAMBDA
  } type;
  union {
    /* Data types */
 /* void unit_v; */
    char bool_v;
    char char_v;
    int int_v;
 /* void null_v; */
    struct { objptr_t car, cdr; } pair_v;
    objptr_t ref_v;
    /* Internal use (but reified) */
    struct { primfunc_t nm; } primfunc_v;
    struct { primfunc_t nm;  objptr_t data1; } primfunc1_v;
    struct { primfunc_t nm;  objptr_t data1, data2; } primfunc2_v;
    struct { hashv_t name;  objptr_t val; } def_v;
 /* void envnull_v; */
    struct { objptr_t def;  objptr_t seq; } env_v;
    struct { hashv_t param;  objptr_t env;  objptr_t body; } userfunc_v;
    /* Continuations */
 /* void cont_finish_v; */
    struct { objptr_t expr;  objptr_t env;  objptr_t cont; } cont_eval_v;
    struct { objptr_t rator;  objptr_t cont; } cont_apply_v;
    struct { objptr_t rand;  objptr_t env;  objptr_t cont; } cont_apply_1_v;
    /* Tasks */
    struct { objptr_t expr;  objptr_t env;  objptr_t cont; } task_eval_v;
    struct { objptr_t rator;  objptr_t rand;  objptr_t cont;} task_apply_v;
    struct { objptr_t rator;  objptr_t rand;  objptr_t env;  objptr_t cont;} task_apply_0_v;
    struct { objptr_t rator;  objptr_t rand;  objptr_t env;  objptr_t cont;} task_apply_1_v;
    struct { objptr_t primus;  objptr_t pares;  objptr_t env;  objptr_t cont;} task_progn_0_v;
    /* Program types */
    hashv_t var_v;
    struct { objptr_t rator, rand; } apply_v;
    struct { objptr_t primus, pares; } progn_v;
    objptr_t quote_v;
    struct { hashv_t param;  objptr_t body; } lambda_v;
  } d;
};

#define CORESIZE 1048676

struct object_s core[CORESIZE];

objptr_t mainprog; /* Set by parser */

objptr_t globenv;

void error (const char *s);

objptr_t allocate (void);

objptr_t a_unit (void);

objptr_t lookup (hashv_t name, objptr_t env);

objptr_t eval (objptr_t expr, objptr_t env, objptr_t cont);

objptr_t step (objptr_t task);

objptr_t invoke (objptr_t cont, objptr_t val);

objptr_t apply_pf_print (objptr_t rand, objptr_t cont);

objptr_t apply_pf_i (objptr_t rand, objptr_t cont);

objptr_t apply_pf_k (objptr_t rand, objptr_t cont);

objptr_t apply_pf_s (objptr_t rand, objptr_t cont);

objptr_t apply_primfunc (primfunc_t fn, objptr_t rand, objptr_t cont);

objptr_t apply_pf1_k (objptr_t data1, objptr_t rand, objptr_t cont);

objptr_t apply_pf1_s (objptr_t data1, objptr_t rand, objptr_t cont);

objptr_t apply_primfunc1 (primfunc_t fn, objptr_t data1,
			  objptr_t rand, objptr_t cont);

objptr_t apply_pf2_s (objptr_t data1, objptr_t data2,
		      objptr_t rand, objptr_t cont);

objptr_t apply_primfunc2 (primfunc_t fn, objptr_t data1, objptr_t data2,
			  objptr_t rand, objptr_t cont);

objptr_t extend_env (hashv_t name, objptr_t val, objptr_t env);

objptr_t apply (objptr_t rator, objptr_t rand, objptr_t cont);

objptr_t build_apply (objptr_t rator, objptr_t rand);

objptr_t build_progn (objptr_t primus, objptr_t pares);

objptr_t build_unit (void);

objptr_t build_var (hashv_t name);

objptr_t build_lambda (hashv_t param, objptr_t body);

objptr_t build_primfunc (primfunc_t nm);

objptr_t build_def (hashv_t name, objptr_t val);

objptr_t build_envnull (void);

objptr_t build_env (objptr_t def, objptr_t seq);

void create_globenv (void);

#endif /* not defined _TUNES_H */
